/*
 * disp.c
 *
 * Read one or more files and display them line-by-line simultaneously
 * originated at givens position on the screen.

 * History:
 *     V0.1: Peter Berger (PEB): January 2001
 *     V0.2: Peter Berger (PEB): January 12 2001
 *       -- Converted to sysio to work better with FIFOs
 *       -- Better ^C handling
 *       -- Converted to use multiplexed i/o via select()
 */
#include <stdio.h>
#include <curses.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#define BSIZE 32
#define MAXFILES 20
#define VERSION "0.1"

static char *version = VERSION" ""$Revision: 1.1 $";
static int stop_flag=0;

struct fstruct {
	char *fname;	// Name of file to display
	int fd;			// ... and corresponding file descriptor
	char *title;	// String to display at top of window
	int x;			// Screen coordinates of origin for
	int y;			// ... this file
	int header;		// Height of header screen.  This many header lines
					// will be displayed followed by a single window for
					// subsequent lines.

	int y_offset;	// Current line offset (starting at 0)
	int x_offset;	// Number of chars on the current output line
} flist[MAXFILES];

static char *cmdname;
static int nfiles;
static fd_set readset, treadset;	// FD descriptor set for select()
void usage();
void catch_sigint( int sig );

void showversion()
{
	printf( "%s: Version  %s\n", cmdname, version );
}

init( int ac, char *av[] )
{
	char *pflag;
	int i, thisfile;

	thisfile=-1;	// -f increments this
	/* Get options */
	while( ac > 0 ) {
		/* handle each flag */
		for( pflag=*av; *pflag != '\0'; pflag++ ) {

			/* switch on flag */
			switch( *pflag ) {

			case '-':
				break;

			case '?':
			case 'h':
				usage();
				return(1);
			case 'v':
				showversion();
				return(1);
			case 't':		// header ("top") height
				if( *(pflag+1) != '\0' ) {
					flist[thisfile].header = atol( pflag+1 );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].header = flist[thisfile].header;
					pflag += strlen(pflag)-1;
				}
				else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -t without value\n",
							cmdname );
						return( 1 );
					}
					flist[thisfile].header = atol( *++av );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].header = flist[thisfile].header;
				}
				break;

			case 'x':
				if( thisfile < 0 ) {
						fprintf( stderr, "%s: -x before -f -- ignoring\n",
							cmdname );
						break;
				}
				if( *(pflag+1) != '\0' ) {
					flist[thisfile].x = atol( pflag+1 );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].x = flist[thisfile].x;
					pflag += strlen(pflag)-1;
				}
				else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -x without value\n",
							cmdname );
						return( 1 );
					}
					flist[thisfile].x = atol( *++av );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].x = flist[thisfile].x;
				}
				break;

			case 'y':
				if( thisfile < 0 ) {
						fprintf( stderr, "%s: -y before -f -- ignoring\n",
							cmdname );
						break;
				}
				if( *(pflag+1) != '\0' ) {
					flist[thisfile].y = atol( pflag+1 );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].y = flist[thisfile].y;
					pflag += strlen(pflag)-1;
				}
				else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -y without value\n",
							cmdname );
						return( 1 );
					}
					flist[thisfile].y = atol( *++av );
					for( i=thisfile+1; i<MAXFILES; ++i )
						flist[i].y = flist[thisfile].y;
				}
				break;

			case 'f':	/* Filename */
				thisfile++;
				if( *(pflag+1) != '\0' ) {
					flist[thisfile].fname = pflag+1;
					pflag += strlen(pflag)-1;
				} else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -f without filename\n",
							cmdname );
						return( 1 );
					}
					flist[thisfile].fname = *++av;
				}
				if( (flist[thisfile].fd=open( flist[thisfile].fname,
				O_RDONLY | O_NONBLOCK )) == -1 ) {
					fprintf( stderr, "%s: Cannot open '%s' - aborting\n",
						cmdname, flist[thisfile].fname );
					return( 1 );
				}
				break;

			case 'T':	/* Window title */
				if( *(pflag+1) != '\0' ) {
					flist[thisfile].title = pflag+1;
					pflag += strlen(pflag)-1;
				} else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -T without title\n",
							cmdname );
						return( 1 );
					}
					flist[thisfile].title = *++av;
				}
				break;

			default:
				usage();
				return( 1 );
			}
		}

		--ac;
		++av;
	}

	nfiles=thisfile+1;
	if( nfiles < 1 ) {
		fprintf( stderr, "%s: no files specified -- aborting.\n",
			cmdname );
		return( 1 );
	}

//	for( i=0; i<nfiles; ++i ) {
//	printf( "fname:'%s', x:%d, y:%d\n", flist[i].fname,  flist[i].x,  flist[i].y );
//	}
//exit(1);



	// ncurses stuff...
	initscr();
	cbreak();
	noecho();
	nonl();
    intrflush(stdscr, FALSE);
    keypad(stdscr, TRUE);
//	timeout(-1);

	return( 0 );


}

main( int ac, char *av[] )
{
	struct timeval tv;
	int pid, sret;
	char buf[BSIZE];
	int i, nc;
	struct sigaction sa;

	cmdname = av[0];
	++av;
	--ac;

	/* catch SIGINT */
	sa.sa_handler = catch_sigint;
	sigemptyset( &sa.sa_mask );
	sa.sa_flags = 0;
	if( sigaction( SIGINT, &sa, NULL ) != 0
	|| sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	/* check for version option */
	if( strcmp( *av, "-version" ) == 0
	|| strcmp( *av, "--version" ) == 0 ) {
		showversion();
		return( 0 );
	}

	if( init( ac, av ) )
		exit( 1 );

	FD_ZERO(&readset);

	for( i=0; i<nfiles; i++ ) {
		FD_SET(flist[i].fd, &readset);
		flist[i].x_offset = flist[i].y_offset = 0;
	}

	tv.tv_sec=1;
	tv.tv_usec=0;
	treadset=readset;
	while( !stop_flag ) {
		tv.tv_sec=1;	// Need to reset these on Linux
		tv.tv_usec=0;
		treadset=readset;
		//sret=select(nfiles, &treadset, (fd_set *)NULL, (fd_set *)NULL, &tv);
		//sret=select(FD_SETSIZE, &treadset, (fd_set *)NULL, (fd_set *)NULL, &tv);
		sret=select(FD_SETSIZE, &treadset, (fd_set *)NULL, (fd_set *)NULL, NULL);

		for( i=0; i<nfiles && sret>0; i++ ) {
			if( FD_ISSET(flist[i].fd, &readset ) ) {
				move( flist[i].y+flist[i].y_offset,
					flist[i].x+flist[i].x_offset );

				if( flist[i].x_offset==0 ) {
					if( flist[i].y_offset==0 && flist[i].title ) {
						addstr( flist[i].title );
						flist[i].x_offset+=strlen(flist[i].title);
					}
				}

				buf[1]='\0';
				nc=read(flist[i].fd,buf,1);
				if( nc > 0 ) {
					if( buf[0] == '\n' ) {
						clrtoeol();
						refresh();
						if( flist[i].y_offset < (flist[i].header-1) )
							flist[i].y_offset++;
						flist[i].x_offset=0;
					} else {
						addstr( buf );
						flist[i].x_offset++;
					}
				}
			}
		}
	}

	endwin();
	for( i=0; i<nfiles; ++i )
		close( flist[i].fd );

}

void usage()
{

	fprintf( stderr,
		"\nusage: %s -f filename [options] [-f filename [options]...]\n", cmdname );
	fprintf( stderr, "\
\n\
where options are:\n\
  -x# and -y# specify the horizontal and vertical screen coordinates at\n\
   which the file contents should be displayed.  Upper left corner is (0,0).\n\
  -t# specifies the number of lines from the top of the file to be displayed\n\
   as a title.  Subsequent lines scroll in a one-line window below this.\n\
  -v specifies the program version number\n\
  -h displays this help message\n\
\n\
   E.g: %s -f /tmp/x -x0 -y0 -t3 -f /tmp/y -x0 -y10 -t3\n", cmdname );
}

/*
*  Catch Signal INT
*/

void catch_sigint( int sig )
{
	stop_flag = 1;
}
