/* svfifo.c
 *
 * Perform what is essentially the equivalent of a continuous "tail -f"
 * that will even work on ephemeral files such as FIFOs.  The idea is
 * to continually update a file with the most recent N lines of a
 * growing file (e.g. a FIFO).
 *
 * History:
 *     V0.1: Peter Berger (PEB): January 2001
 */
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>

#define MAXLINELEN 512
#define VERSION "0.1"
#define MAXLINES 10
static char *version = VERSION" ""$Revision: 1.1 $";

static char *cmdname;
static char *in_fname, *out_fname;
static int in_fd, out_fd;
static int maxlines=MAXLINES;	// Number of lines to save
static int nlines=0;	// Number of lines currently in our list
static int stop_flag=0;

void usage();
void catch_sigint( int sig );

// Linked list of remembered lines
struct llist {
	char *line;	// Data bytes
	int len;	// Number of data bytes
	struct llist *next;
} *llist_head, *llist_tail;

void showversion()
{
	printf( "%s: Version  %s\n", cmdname, version );
}

init( int ac, char *av[] )
{
	char *pflag;
	int i, thisfile;

	llist_head=llist_tail=NULL;
	in_fname=out_fname=NULL;

	/* Get options */
	while( ac > 0 ) {
		/* handle each flag */
		for( pflag=*av; *pflag != '\0'; pflag++ ) {

			/* switch on flag */
			switch( *pflag ) {

			case '-':
				break;

			case '?':
			case 'h':
				usage();
				return(1);
			case 'v':
				showversion();
				return(1);

			case 'i':	/* Input Filename  -- usually a FIFO */
				if( *(pflag+1) != '\0' ) {
					in_fname = pflag+1;
					pflag += strlen(pflag)-1;
				} else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -i without filename\n",
							cmdname );
						return( 1 );
					}
					in_fname = *++av;
				}
				break;

			case 'o':	/* Output filename */
				if( *(pflag+1) != '\0' ) {
					out_fname = pflag+1;
					pflag += strlen(pflag)-1;
				} else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -o without filename\n",
							cmdname );
						return( 1 );
					}
					out_fname = *++av;
				}
				break;

			case 'n':
				if( *(pflag+1) != '\0' ) {
					maxlines = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --ac <= 0 ) {
						fprintf( stderr, "%s: -n without value\n",
							cmdname );
						maxlines=MAXLINES;
						return( 1 );
					}
					maxlines = atol( *++av );
				}
				break;

			default:
				usage();
				return( 1 );
			}
		}

		--ac;
		++av;
	}

	if( !strlen(in_fname) || !strlen(in_fname) ) {
		fprintf( stderr, "%s: Missing filename (need -i and -o).\n",
			cmdname );
		return( 1 );
	}
	//if( (in_fd=open( in_fname, O_RDONLY | O_NONBLOCK )) == -1 )
	if( (in_fd=open( in_fname, O_RDONLY )) == -1 )
	{
		fprintf( stderr, "%s: Cannot open input file: '%s' - aborting\n",
			cmdname, in_fname );
		return( 1 );
	}

	return( 0 );
}

main( int ac, char *av[] )
{
	char buf[MAXLINELEN], *bufpt, *lpt;
	int i, c, h, llen;
	int nc;
	struct sigaction sa;

	cmdname = av[0];
	++av;
	--ac;

	/* catch SIGINT */
	sa.sa_handler = catch_sigint;
	sigemptyset( &sa.sa_mask );
	sa.sa_flags = 0;
	if( sigaction( SIGINT, &sa, NULL ) != 0
	|| sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}
	if( sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	/* check for version option */
	if( strcmp( *av, "-version" ) == 0
	|| strcmp( *av, "--version" ) == 0 ) {
		showversion();
		return( 0 );
	}

	if( init( ac, av ) )
		exit( 1 );

	update();	// Truncate the output file

	bufpt=buf;
	for( bufpt=buf, llen=0;!stop_flag; ) {
		nc=read(in_fd,bufpt,1);

		if( nc > 0 ) {
			llen += nc;
			if( *bufpt=='\n' || llen >= MAXLINELEN ) {	// End of a line
				if( (lpt = (char *)malloc( llen )) == NULL )
					fprintf( stderr, "%s: malloc of line space failed\n",
						cmdname, strerror( errno ) );
				else {
					memcpy( lpt, buf, llen );
					appendline( lpt, llen );	// Add to linked list
					update();			// Update the output file
					llen = 0;
					bufpt=buf;
				}
			} else
				bufpt++;
		}


	}

	close( in_fd );		// Note: output file closed in update()

}

/*
*  Catch Signal INT
*/

void catch_sigint( int sig )
{
	stop_flag = 1;
}

void usage()
{

	fprintf( stderr,
		"\nusage: %s -i fname -o fname [-n maxlines] ...]\n", cmdname );
	fprintf( stderr, "\
\n\
where options are:\n\
  -i input filename\n\
  -o output filename\n\
  -n number of lines to preserve\n\
  -v display program version number and exit\n\
  -h display this help message and exit\n", cmdname );
}


/* Add line to linked list and delete early lines if
 * we've got more than we want.
 */
appendline( char *lpt, int len )
{
	struct llist *llist_pt;
	
	if( (llist_pt = (struct llist *)malloc( sizeof(struct llist) )) == NULL )
		fprintf( stderr, "%s: malloc of list element space failed\n",
				cmdname, strerror( errno ) );

	llist_pt->line = lpt;
	llist_pt->next = NULL;
	llist_pt->len = len;

	if( llist_head == NULL )	// First record?
		llist_head = llist_pt;

	if( llist_tail )
		llist_tail->next = llist_pt;
	llist_tail=llist_pt;

	if( nlines < maxlines )	// Still accumulating
		++nlines;
	else {					// Too many lines, drop the oldest
		llist_pt = llist_head->next;
		free( llist_head->line );
		free( llist_head );
		llist_head = llist_pt;
	}
	
}


update() {			// Update the output file
	struct llist *pt;

	pt=llist_head;

	if( (out_fd=open( out_fname,
	O_WRONLY | O_CREAT | O_TRUNC,
	S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH)) == -1 ) {
		fprintf( stderr, "%s: Cannot open output file: '%s' - aborting\n",
			cmdname, out_fname );
		return( 1 );
	}


	while( pt != NULL ) {
		if( write( out_fd, pt->line, pt->len ) != pt->len )
			fprintf( stderr, "%s: write failed, %s\n",
				cmdname, strerror( errno ) );
		pt=pt->next;
	}

	close( out_fd );

}
