/* modem.c
 * Test modem signals.
 *
 * History:
 *     V0.1: Peter Berger (PEB): February 2001
 *         -- First version.
 *     V0.2: PEB: March 12, 2001
 *         -- Added automatic checks for expected behavior.
 *         -- Added to help message.
 *         -- Added "-m", manual mode.
 *         -- Version is now 0.2.
 */
/* includes */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>

/* functions */
void catch_sigint( int sig );
void usage();
int get_tios( int *fd, char *ttyname, struct termios *tios );
int set_tios( int *fd, char *ttyname, struct termios *tios );

#define VERSION "0.2"
#define NAMELEN 128
#define BUF_SIZE 128

/* globals */
static char *cmdname;
static char in_ttyname[NAMELEN];
static char out_ttyname[NAMELEN];
static char *version = VERSION" ""$Revision: 1.1 $";

static int stop_flag = 0;


void showversion()
{
	printf( "%s: Version  %s\n", cmdname, version );
}


void display( char *mes, int status )
{
	printf( "%s: ", mes );
	printf( "DTR:%3s, ", status & TIOCM_DTR? "ON" : "OFF" );
	printf( "RTS:%3s, ", status & TIOCM_RTS? "ON" : "OFF" );
	printf( "CTS:%3s, ", status & TIOCM_CTS? "ON" : "OFF" );
	printf( "DSR:%3s, ", status & TIOCM_DSR? "ON" : "OFF" );
	printf( "CD:%3s, ", status & TIOCM_CD?  "ON" : "OFF" );
	printf( "RI:%3s\n", status & TIOCM_RI?  "ON" : "OFF" );

}

void next( int manual ) {
	int c;
	if( manual ) {
		printf( "Press Enter to continue..." );
		c = getchar();
	}
	printf( "\n" );
}

void uerror( char *mes ) {
	fprintf( stderr, "ERROR: %s", mes );
}


/*
*  Main
*/

int main( int argc, char *argv[] )
{

	int ret, out_fd, in_fd, manual, error,
		init_out_status, init_in_status, out_status, in_status;
	long l, fails;
	char *ttyname,*pflag;
	struct sigaction sa;
	char *pattern_pt, *testname;


	/* get command name */
	cmdname = argv[0];
	--argc;
	++argv;

	/* catch SIGINT */
	sa.sa_handler = catch_sigint;
	sigemptyset( &sa.sa_mask );
	sa.sa_flags = 0;
	if( sigaction( SIGINT, &sa, NULL ) != 0
	|| sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	/* check args */
	if( argc < 1 ) {
		usage( );
		return( 0 );
	}

	/* check for version option */
	if( strcmp( *argv, "-version" ) == 0
	|| strcmp( *argv, "--version" ) == 0 ) {
		showversion();
		return( 0 );
	}


	error = 0;
	manual = 0;	/* Do not pause for keypresses between transitions */
	strcpy( in_ttyname, "" );
	/* get options */
	while( argc > 0 ) {

		/* handle each flag */
		for( pflag=*argv; *pflag != '\0'; pflag++ ) {

			/* switch on flag */
			switch( *pflag ) {

			case '-':
				break;

			case '?':
			case 'h':
				usage();
				return(0);

			case 'v':
				showversion();
				return(0);

			case 'm':
				manual = 1;
				break;

			case 'i':			/* Input file name */
				if( *(pflag+1) != '\0' ) {
					strncpy( in_ttyname, pflag+1, NAMELEN-1 );
				} else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -i without input file name\n",
							cmdname );
						return( 1 );
					} else
						strncpy( in_ttyname, *++argv, NAMELEN-1 );
				}
				break;

			case 'o':			/* Output file name */
				if( *(pflag+1) != '\0' ) {
					strncpy( out_ttyname, pflag+1, NAMELEN-1 );
				} else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -o without input file name\n",
							cmdname );
						return( 1 );
					} else
						strncpy( out_ttyname, *++argv, NAMELEN-1 );
				}
				break;

			default:
				fprintf( stderr, "%s: bad flag '%c'\n", cmdname, *pflag );
				usage();
				return( 1 );
				break;

			}

		}

		--argc;
		++argv;

	}


	/* open ports */
	if( (in_fd=open( in_ttyname, O_RDWR ) ) < 0 ) {
		fprintf( stderr, "%s: open failed on %s, %s\n",
			cmdname, in_ttyname, strerror( errno ) );
		return( 1 );
	}
	if( (out_fd=open( out_ttyname, O_RDWR ) ) < 0 ) {
		fprintf( stderr, "%s: open failed on %s, %s\n",
			cmdname, out_ttyname, strerror( errno ) );
		return( 1 );
	}

	/* Get initial status */
	ioctl( out_fd, TIOCMGET, &init_out_status );
	ioctl( in_fd, TIOCMGET, &init_in_status );
	out_status = init_out_status;
	in_status = init_in_status;

	printf( "Initial port settings:\n" );
	display( "output port", out_status );
	display( " input port", in_status );
	next(manual);


	out_status &= ~(TIOCM_DTR | TIOCM_RTS | TIOCM_CTS | TIOCM_DSR | TIOCM_CD | TIOCM_RI );
	/* Initially set in and out ports to this known, cleared state */
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( in_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After clearing all modem signals:\n" );
	display( "output port", out_status );
	display( " input port", in_status );
	next(manual);

	out_status |= TIOCM_DTR;
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After setting DTR on output port:\n");
	display( "output port", out_status );
	display( " input port", in_status );
	if( (in_status & (TIOCM_DSR|TIOCM_CD)) != (TIOCM_DSR|TIOCM_CD) ) {
		uerror(
		"Setting DTR on output port did not set DSR/DCD on input port\n" );
printf( "in_status: 0x%x, TIOCM_DSR:0x%x, TIOCM_CD:0x%x, in_status & (TIOCM_DSR|TIOCM_CD):0x%x, TIOCM_DSR|TIOCM_CD:0x%x\n",
in_status, TIOCM_DSR, TIOCM_CD, in_status & (TIOCM_DSR|TIOCM_CD), TIOCM_DSR|TIOCM_CD );
		error++;
	}
	next(manual);


	out_status |= TIOCM_RTS;
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After setting RTS on output port:\n");
	display( "output port", out_status );
	display( " input port", in_status );
	if( !(in_status & TIOCM_CTS) ) {
		uerror(
		"Setting RTS on output port did not set CTS on input port\n" );
		error++;
	}
	next(manual);

	out_status &= ~TIOCM_DTR;
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After clearing DTR on output port:\n");
	display( "output port", out_status );
	display( " input port", in_status );
	if( in_status & (TIOCM_DSR|TIOCM_CD) ) {
		uerror(
		"Clearing DTR on output port did not clear DSR/DCD on input port\n" );
		error++;
	}
	next(manual);

	out_status &= ~TIOCM_RTS;
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After clearing RTS on output port:\n");
	display( "output port", out_status );
	display( " input port", in_status );
	if( in_status & TIOCM_CTS ) {
		uerror(
		"Clearing RTS on output port did not clear CTS on input port\n" );
		error++;
	}
	next(manual);

	/* Restore to initial settings */
	ioctl( out_fd, TIOCMSET, &init_out_status );
	ioctl( in_fd, TIOCMSET, &init_in_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );
	printf( "After restoring initial settings on both ports:\n");
	display( "output port", out_status );
	display( " input port", in_status );
	next(manual);

	/* close port */
	close( out_fd );
	close( in_fd );
	return( error );

}



/*
*  Catch Signal INT
*/

void catch_sigint( int sig )
{
	stop_flag = 1;
printf( "catch_sigint: caught sig:%d\n", sig );
}


/*
*  Usage
*/

void usage()
{

	fprintf( stderr, "\nusage: %s [options] -i ttyname -o ttyname\n", cmdname );

	fprintf( stderr,
	"Where -i and -o specify the input and output ports, and options are:\n\
        -v     Display version number and quit\n\
        -h     Display this usage message\n\
        -m     \"manual\": pause for keypress following each transition\n" );

	fprintf( stderr, "\nNotes:\n\
       1. The input and output ports must be wired:\n\
          RxD  -- TxD          TxD     -- RxD\n\
          RTS  -- CTS          CTS     -- RTS\n\
          DTR  -- DCD/DSR      DCD/DSR -- DTR\n\
       2. The input and output ports can be one and the same if connected\n\
          with an external loopback plug wired as above.\n" );



}

