/* utest.c
 * Send user-specified or randomly generated data from one port to another
 * and optionally check to see if it arrives intact.
 *
 * History:
 *     V0.1: Peter Berger (PEB): December 2000
 *         -- First version
 *     V0.2: PEB: 7 Jan 2001
 *         -- Added fail counter to status message
 *         -- Change test loop counter to start counting at 1 (was 0)  
 *         -- Added 921600 baud rate
 *     V0.3: PEB: 10 Jan 2001
 *         -- Added MAXTIMEOUTS
 *         -- Support for infinite test runs (by setting "cnt' to -1)
 *         -- Added initial display of test run specifications
 *         -- Add output only test (no input testing if no -i flag specified)
 *         -- Use setvbuf() to set stdout to be line buffered for more
 *            intuitive viewing via "disp"
 *     V0.4: PEB: 17 Jan 2001
 *         -- Add bidirectional Cross_Talk()
 *         -- Fixed bug with user-specified output pattern
 *     V0.5: PEB: 20 Feb 2001
 *         -- Fixed usage message
 *     V0.6: PEB: 12 March 2001
 *         -- Added opt_* flags so options can be specified before or
 *            after the -i and -o specifications.  Previously options
 *            such as baudrate had to be specified after the port name
 *            since the in and out_tios structures were reset at option
 *            parsing time.
 *         -- Added modem line tests ala tmodem.c.
 *         -- Added call to DIGI_TIMERIRQ to enable/disable RS_TIMER use.
 *     V0.7: PEB: 19 March 2001
 *         -- Edited "Pass rate" message to prevent interpretation confusion.
 *     V0.8: PEB: 4 April 2001
 *         -- Removed some now unused code, some of which caused compilation
 *            problems when moving between 2.2 and 2.4 kernels.
 *     V0.9: PEB: 23 April 2001
 *         -- Added "-L" flag to enable internal UART loopback.
 *     V0.10: PEB: 4 May 2001
 *         -- Added "fd" argument to usage(FILE *out) so "utest -h" can now
 *            go to stdout from which it can be paged with "more" without
 *            requiring stderr redirection.
 *         -- Removed a leftover debug printf for opt_L.
 *     V0.11: PEB: 24 May 2001
 *         -- Add kbhit() and "-A" command line option to signal that any
 *            keypress should abort tests in progress.
 *     V0.12: PEB: 29 May 2001
 *         -- Add "-R" option to display summary report at end of run.
 *         -- Added message on user abort.
 *     V0.13: PEB and Al Borchers (ATB): 8 June 2001
 *         -- Changed "CrossTalk" to "Port-to-Port".
 *         -- Changed format of "-R" summary report.
 *         -- Added O_NONBLOCK on reads, so that large tests do not hang.
 *     V1.0: PEB: 9 June 2001
 *         -- Flush buffers on close.
 *     V1.1: PEB: 12 September 2001
 *         -- Vantive #7223: when the user interrupts a series of tests via
 *            Ctl/C, do not include the final (interrupted) test in the
 *            failure count and statistics report.  Also, increment the
 *            test index number when reporting interrupted tests since user
 *            counts start at 1 (rather than 0).
 *         -- Add -T option to allow user to leave RS_TIMER enabled during
 *            tests.  This is useful for tracking down IRQ configuration
 *            problems.
 *     V1.2: Steve Dunlop (SJD): May 2003
 *         -- Avoid embedding XON or XOFF characters in random data streams
 *            during tests where software flow control is enabled.
 */
/* includes */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <linux/types.h>
#include <curses.h>
#include <limits.h>

//#include <term.h>

#include "digi.h"
#include "utest.h"

/* typedefs */
typedef struct {
	long baud_num;
	int baud_flag;
} Baud_Table_Entry;


/* functions */
void got_quit( void );
void catch_int( int sig );
void usage(FILE *out);
int get_tios( int *fd, char *ttyname, struct termios *tios );
int set_tios( int *fd, char *ttyname, struct termios *tios );
int Cross_Talk( long tno, int in_fd, int out_fd, int Xon );
int Output_Only( long tno, int out_fd );

#define VERSION "1.2"
#define NAMELEN 128
#define BUF_SIZE 128

#define DEFAULT_BAUDRATE 9600
#define DEFAULT_CNT 1
#define DEFAULT_PATLEN 10

#define XON_CHAR ('q' & 0x1f )
#define XOFF_CHAR ('s' & 0x1f )

// Maximum number of read timeouts before a given test is aborted 
#define MAXTIMEOUTS 5


/* globals */
static char *cmdname;
static char in_ttyname[NAMELEN];
static char out_ttyname[NAMELEN];
static char *pattern=NULL;
static long baud=DEFAULT_BAUDRATE,
	cnt=DEFAULT_CNT,
	patlen=DEFAULT_PATLEN;
static char outbuf[BUFSIZ];	// We'll remap stdout using setvbuf()

static char *version = VERSION" ""$Revision: 1.3 $";

static int stop_flag = 0;
static int bidirect = 0;	// Set if CrossTalk tests should go both ways
static int anykey_abort = 0;	// Set if any keypress should abort tests
static int double_error = 0;

static Baud_Table_Entry translate_baud[] = {
	{ 0, B0 },
	{ 50, B50 },
	{ 75, B75 },
	{ 110, B110 },
	{ 134, B134 },
	{ 150, B150 },
	{ 200, B200 },
	{ 300, B300 },
	{ 600, B600 },
	{ 1200, B1200 },
	{ 1800, B1800 },
	{ 2400, B2400 },
	{ 4800, B4800 },
	{ 9600, B9600 },
	{ 19200, B19200 },
	{ 38400, B38400 },
	{ 57600, B57600 },
	{ 115200, B115200 },
	{ 230400, B230400 },
	{ 460800, B460800 },
	{ 921600, B921600 },
	{ -1L, 0 }
};


static struct termios initial_tios, new_tios;
void init_kb()
{
	tcgetattr(0,&initial_tios);
	new_tios = initial_tios;
	new_tios.c_lflag &= ~ICANON;
	new_tios.c_lflag &= ~ECHO;
	new_tios.c_cc[VMIN] = 0;	// Return even without data
	new_tios.c_cc[VTIME] = 0;
	tcsetattr(0,TCSANOW, &new_tios);


}
void close_kb()
{
	tcsetattr(0,TCSANOW, &initial_tios);
}

/* kbhit:  Has user pressed a key?
 */
int kbhit()
{
	int ch;

	if( read(0,&ch,1) ) {
		got_quit();
		return( 1 );
	} else
		return( 0 );
}

void uerror( char *mes )
{
	char *leader = "     MODEMLINE ERROR: ";
	fprintf( stdout, "%s%s", leader, mes );
	if( double_error )
		fprintf( stderr, "%s%s", leader, mes );
}

int modemtests( int in_fd, int out_fd )
{

	int init_out_status, init_in_status, out_status, in_status, error;

	/* Get initial status */
	ioctl( out_fd, TIOCMGET, &init_out_status );
	ioctl( in_fd, TIOCMGET, &init_in_status );
	out_status = init_out_status;
	in_status = init_in_status;

	out_status &= ~(TIOCM_DTR | TIOCM_RTS | TIOCM_CTS | TIOCM_DSR | TIOCM_CD | TIOCM_RI );
	/* Initially set in and out ports to this known, cleared state */
	ioctl( out_fd, TIOCMSET, &out_status );
	ioctl( in_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	out_status |= TIOCM_DTR;
	ioctl( out_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	/*
	 * This original test is somewhat flawed.
	 *
	 * It assumes the customer has a cable that wires DTR to both DSR and DCD.
	 * This is rather unusual.
	 *
	 * Most customers have a cable that wires DTR to either DSR or DCD.
	 * Because of this, lets be happy if EITHER DSR or DCD goes up...
	 */
//	if( (in_status & (TIOCM_DSR|TIOCM_CD)) != (TIOCM_DSR|TIOCM_CD) ) {
	if( (in_status & (TIOCM_DSR|TIOCM_CD)) == 0 ) {
		uerror(
		"Setting DTR on output port did not set DSR/DCD on input port\n" );
		error++;
	}

	out_status |= TIOCM_RTS;
	ioctl( out_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	if( !(in_status & TIOCM_CTS) ) {
		uerror(
		"Setting RTS on output port did not set CTS on input port\n" );
		error++;
	}

	out_status &= ~TIOCM_DTR;
	ioctl( out_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	if( in_status & (TIOCM_DSR|TIOCM_CD) ) {
		uerror(
		"Clearing DTR on output port did not clear DSR/DCD on input port\n" );
		error++;
	}
	out_status &= ~TIOCM_RTS;
	ioctl( out_fd, TIOCMSET, &out_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	if( in_status & TIOCM_CTS ) {
		uerror(
		"Clearing RTS on output port did not clear CTS on input port\n" );
		error++;
	}

	/* Restore to initial settings */
	ioctl( out_fd, TIOCMSET, &init_out_status );
	ioctl( in_fd, TIOCMSET, &init_in_status );

	ioctl( out_fd, TIOCMGET, &out_status );
	ioctl( in_fd, TIOCMGET, &in_status );

	return( error );
}

/* get_pattern
 * Convert user-specified pattern (a hex-coded ASCII string) to the
 * byte array and store it in malloc'd space pointed to by the global
 * pointer "pattern".
 *
 * Return 0 on success, 1 on error
 */
int get_pattern( char *cpattern )
{
	char *cpt, *ipt, bbuf[3], byte;
	int len, mlen;
int i;

	cpt = cpattern;	// Character (ASCII) input
	ipt = pattern;	// Integer array output
	len=strlen(cpt);

	if( strncasecmp( cpt, "0x", 2 ) ) {
		fprintf( stderr, "get_pattern missing leading \"0x\".  Ignoring pattern spec.\n" );
		return( 1 );
	}
	cpt += 2;
	len -= 2;

	if( len <= 0 ) {
		fprintf( stderr, "get_pattern failed to parse pattern '%s'\n", cpattern );
		return( 1 );
	}

		
	mlen = (len / 2);	// Two ASCII input chars per integer byte output
	if( (pattern=malloc(mlen)) == NULL ) {
		fprintf( stderr, "get_pattern malloc failed\n" );
		return( 1 );
	}

	ipt=pattern;
	while( len >= 2 ) {
		strncpy( bbuf, cpt, 2 );
		byte = strtol( bbuf, NULL, 16 );
		*ipt++ = byte;
		cpt += 2;
		len -= 2;
	}
	patlen=mlen;

	return( 0 );

}

void showversion()
{
	printf( "%s: Version  %s\n", cmdname, version );
}

show_result( FILE *out,long l, int ret, long fails ) 
{
	fprintf( out, "     Test #%ld", l+1 );
	if( cnt > 0 )
		fprintf( out,"/%ld", cnt );

 	fprintf( out, " '%s'. (%ld Pass/%ld Fail; Pass rate:%.1f%%)\n",
		ret? "Failed" : "Passed", l+1-fails, fails,
		(100.0*(l+1-fails))/(l+1) );
}


void banner( char *testname, int opt_L )
{
	printf("\
%s specifications:\n\
     Using %d byte pattern of %s data\n",
	testname,
	patlen, pattern == NULL? "Random-generated" : "User-specified" );
	printf("\
     Number of test iterations to run: " );
	if( cnt > 0 )
		printf( "%ld", cnt );
	else
		printf( "infinite loop" );
	if( opt_L )
		printf( "\n\
     using internal UART loopback mode" );
	printf( "\n" );
	if( bidirect && in_ttyname[0] )
		printf("\
     Bidirectional on ports: %s and %s, Baudrate:%ld\n",
	in_ttyname[0] ? in_ttyname : "NONE", out_ttyname, baud );
	else
		printf("\
     Input port: %s, Output port: %s, Baudrate:%ld\n",
	in_ttyname[0] ? in_ttyname : "NONE", out_ttyname, baud );
}

void summary( char *testname, int opt_L, int opt_H, int opt_S, long l, long fails, int stop_flag )
{
	char *offset="     ";

	printf( "\nTest Results Summary:\n" );
	printf( "%sCompleted %d ", offset, l );

	if( opt_L )
		printf( "internal loopback" );
	else if( !strcmp( in_ttyname, out_ttyname ) )
		printf( "external loopback" );
	else
		printf( "port-to-port" );

	printf( " test%s", l==1? "" : "s" );

	if( opt_L || !strcmp( in_ttyname, out_ttyname) )
    	printf( " on port %s", out_ttyname );
	else
		printf( " from port %s to %s", out_ttyname, in_ttyname );

	printf( "\n%s", offset );
	if( bidirect )
		printf( "and back again " );

	printf( "at %ld baud using %d bytes of %s data\n",
		baud, patlen, pattern == NULL? "Random-generated" : "User-specified" );
	printf( "%swith ", offset );

	if( opt_H || opt_S ) {
		if( opt_H ) {
			printf( "hardware%s", opt_S? " and ": "" );
		}
		if( opt_S )
			printf( "software" );
	} else
		printf( "no" );
	printf( " flow control.  %s UART registers\n",
		l==1? "This test exercises" : "These tests exercise" );
	printf( "%sand FIFO memory, baud rate configuration%s",
		offset, opt_H? ", hardware flow control\n" : "" );
	printf( "%sand adapter interrupts.\n", opt_H? offset: " " );

	printf( "%sOf the %d test%s, %d passed, %d failed; Pass rate:%.1f%%\n",
		offset, l, l==1? "": "s", l-fails, fails, (100.0*(l-fails))/l );

	if( stop_flag )
		printf( "\n%sNote: Testing terminated early due to interrupt from user.\n",
		offset );
}

/*
*  Main
*/

int main( int argc, char *argv[] )
{

	int ret,in_fd,out_fd, do_modemtests, modem_fails, ctlfd;
	long l, fails;
	char *ttyname,*pflag;
	Baud_Table_Entry *pbaud;
	struct termios in_tios, out_tios;
	struct sigaction sa;
	char *pattern_pt, *testname;

	/* Keep track of which termios options should be set after ports have
	 * been opened */
	Baud_Table_Entry *opt_b = NULL;
	int	opt_R = 0;
	int	opt_L = 0;
	int	opt_e = 0;
	int opt_S = 0;
	int opt_H = 0;
	do_modemtests = 0;
	double_error = 0;


	/* get command name */
	cmdname = argv[0];
	--argc;
	++argv;

	/* catch SIGINT */
	sa.sa_handler = catch_int;
	sigemptyset( &sa.sa_mask );
	sa.sa_flags = 0;
	if( sigaction( SIGINT, &sa, NULL ) != 0
	|| sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	/* check args */
	if( argc < 1 ) {
		usage(stderr);
		return( 0 );
	}

	/* check for version option */
	if( strcmp( *argv, "-version" ) == 0
	|| strcmp( *argv, "--version" ) == 0 ) {
		showversion();
		return( 0 );
	}


	strcpy( in_ttyname, "" );
	strcpy( out_ttyname, "" );
	/* get options */
	while( argc > 0 ) {

		/* handle each flag */
		for( pflag=*argv; *pflag != '\0'; pflag++ ) {

			/* switch on flag */
			switch( *pflag ) {

			case '-':
				break;

			case '?':
			case 'h':
				usage(stdout);
				return(0);

			case 'm':
				do_modemtests = 1;
				break;

			case 'v':
				showversion();
				return(0);

			case 'B':
				bidirect=1;
				break;

			case 'A':
				anykey_abort=1;
				break;

			case 'H':
				opt_H = 1;
				//in_tios.c_cflag |= CRTSCTS;
				//out_tios.c_cflag |= CRTSCTS;
				break;

			case 'S':
				opt_S = 1;
				//in_tios.c_iflag |= (IXON|IXOFF);
				//out_tios.c_iflag |= (IXON|IXOFF);
				break;

			case 'e':
				opt_e = 1;
				//in_tios.c_lflag |= ECHO;
				//out_tios.c_lflag |= ECHO;
				break;

			case 'E':
				double_error = 1;
				break;

			case 'i':			/* Input file name */
				if( *(pflag+1) != '\0' ) {
					strncpy( in_ttyname, pflag+1, NAMELEN-1 );
				} else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -i without input file name\n",
							cmdname );
						return( 1 );
					} else
						strncpy( in_ttyname, *++argv, NAMELEN-1 );
				}
				if( get_tios( &in_fd, in_ttyname, &in_tios ) ){
					fprintf( stderr, "%s: get_tios failed, %s\n",
					cmdname, strerror( errno ) );
					return( 1 );
				}
				break;

			case 'o':			/* Output file name */
				if( *(pflag+1) != '\0' ) {
					strncpy( out_ttyname, pflag+1, NAMELEN-1 );
				} else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -o without output file name\n",
							cmdname );
						return( 1 );
					} else
						strncpy( out_ttyname, *++argv, NAMELEN-1 );
				}
				if( get_tios( &out_fd, out_ttyname, &out_tios ) ){
					fprintf( stderr, "%s: get_tios failed, %s\n",
					cmdname, strerror( errno ) );
					return( 1 );
				}
				break;


			case 'b':
				if( *(pflag+1) != '\0' ) {
					baud = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -b without baud rate\n",
							cmdname );
						return( 1 );
					}
					baud = atol( *++argv );
				}

				/* translate baud rate */
				for( pbaud=translate_baud; pbaud->baud_num != -1; pbaud++ ) {
					if( baud == pbaud->baud_num ) {
						opt_b = pbaud;
						//in_tios.c_cflag &= ~CBAUD;
						//in_tios.c_cflag |= pbaud->baud_flag;
						//out_tios.c_cflag &= ~CBAUD;
						//out_tios.c_cflag |= pbaud->baud_flag;
						break;
					}
				}

				/* baud rate not found */
				if( pbaud->baud_num == -1 ) {
					fprintf( stderr, "%s: unknown baud rate %d\n", cmdname,
						baud );
					return( 1 );
				}

				break;

			case 'n':
				if( *(pflag+1) != '\0' ) {
					cnt = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -c without repetition count\n",
							cmdname );
						return( 1 );
					}
					cnt = atol( *++argv );
				}
				break;
			case 'L':
				opt_L = 1;	/* Enable internal UART loopback mode */
				break;
			case 'R':
				opt_R = 1;	/* Display summary report at end */
				break;
			case 'l':
				if( *(pflag+1) != '\0' ) {
					patlen = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -l without pattern length\n",
							cmdname );
						return( 1 );
					}
					if( pattern ) {
						fprintf( stderr, "%s: pattern length set implicitly by -P option, ignoring -l option\n",
							cmdname );
							++argv;
					} else
						patlen = atol( *++argv );
				}
				break;

			case 'P':	/* Use this pattern instead of the random default */
				if( *(pflag+1) != '\0' ) {
					pattern_pt = pflag+1;
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -P without pattern\n",
							cmdname );
						return( 1 );
					}
					pattern_pt = *++argv;
				}
				if( get_pattern( pattern_pt ) )
					pattern_pt = NULL;
				break;


			default:
				fprintf( stderr, "%s: bad flag '%c'\n", cmdname, *pflag );
				usage(stderr);
				return( 1 );
				break;

			}

		}

		--argc;
		++argv;

	}


	/* Any user-specified options to override current termios settings? */
	if( opt_b ) {
		in_tios.c_cflag &= ~CBAUD;
		in_tios.c_cflag |= opt_b->baud_flag;
		out_tios.c_cflag &= ~CBAUD;
		out_tios.c_cflag |= opt_b->baud_flag;
	}
	if( opt_e ) {
			in_tios.c_lflag |= ECHO;
			out_tios.c_lflag |= ECHO;
	}
	if( opt_H ) {
			in_tios.c_cflag |= CRTSCTS;
			out_tios.c_cflag |= CRTSCTS;
	}
	if( opt_S ) {
			in_tios.c_iflag |= (IXON|IXOFF);
			out_tios.c_iflag |= (IXON|IXOFF);
	}

	srand((unsigned)time(NULL));	// Seed the random number generator
	if( setvbuf( stdout,outbuf,_IOLBF, BUFSIZ ) ) {
		fprintf( stderr, "%s: setvbuf failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	if( in_ttyname[0] ) {
		if(  set_tios( &in_fd, in_ttyname, &in_tios ) ) {
			fprintf( stderr, "%s: set_tios failed, %s\n",
			cmdname, strerror( errno ) );
			return( 1 );
		}
		if( opt_L ) {
			if( (ret=ioctl( in_fd, DIGI_LOOPBACK, &opt_L )) < 0 ) {
				fprintf( stderr, "%s: DIGI_LOOPBACK failed, ret=%d, %s\n", cmdname,
				ret, strerror( errno ) );
			}
		}
	}

	if( set_tios( &out_fd, out_ttyname, &out_tios ) ) {
		fprintf( stderr, "%s: set_tios failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	if( opt_L ) {
		if( (ret=ioctl( out_fd, DIGI_LOOPBACK, &opt_L )) < 0 ) {
			fprintf( stderr, "%s: DIGI_LOOPBACK failed, ret=%d, %s\n", cmdname,
			ret, strerror( errno ) );
		}
	}

	if( in_ttyname[0] )
		testname="Port-to-Port";
	else
		testname="Output-Only";

	banner( testname, opt_L );

	init_kb();
	fails=0;
	// If cnt is negative, loop forever
	for( l=0; !stop_flag && (cnt>0? l<cnt: 1); ++l )
	{
		if( in_ttyname[0] ) {
			ret=Cross_Talk( l, in_fd, out_fd, opt_S );
			if( bidirect )
				ret+=Cross_Talk( l, out_fd, in_fd, opt_S );
		} else
			ret=Output_Only( l, out_fd );
		if( !stop_flag ) {	/* Skip reporting after user abort (Ctl/C) */
			if( ret )
				fails++;
			show_result( stdout,l,ret,fails );
			if( ret && double_error )	// Log failures to stderr
				show_result( stderr,l,ret,fails );
	
			if( do_modemtests )
				modem_fails += modemtests( in_fd, out_fd );
	
			// User keypress?
			if( anykey_abort )
				(void)kbhit();
		}
	}

	/* If user aborted the test, don't included the aborted iteration
	 * in summary */
	if( stop_flag )
		l--;

	if( opt_R ) {
		summary( testname, opt_L, opt_H, opt_S, l, fails, stop_flag );
	}

	/* close port */
	close_kb();
	if( tcflush( in_fd, TCIOFLUSH ) || tcflush( out_fd, TCIOFLUSH ) )
		fprintf( stderr, "%s: Buffer flush failed.\n", cmdname );
	close( in_fd );
	close( out_fd );
	if( pattern )
		free( pattern );

	return( 0 );

}

/* rdata
 * Return a bits-size byte of random data
 */
unsigned char rdata( int bits )
{
	unsigned int mask, ret;

	mask = (1 << bits) - 1;
	return( (unsigned char)(rand() & mask ) );
}

int Cross_Talk( long tno, int in_fd, int out_fd, int Xon )
{
	int err, i, flags, ret, write_chars, read_chars, freeout, timeouts, ltno;
	char *out_buf, *in_buf;

	ltno = tno+1;	/* For users start counting at 1 not 0 */
	/* If user did not specify a pattern, malloc space and
	 * generate a random pattern */
	freeout = 0;
	if( pattern == NULL ) {
		freeout = 1;
		if( (out_buf = malloc( patlen )) == NULL ) {
			fprintf( stderr, "Cross_Talk out_buf malloc failed, test #%ld\n",
			ltno );
			return( 1 );
		}
		for( i=0; i<patlen; ++i ) {
			do {	// if using software flow control, roll dice again
					// whenever XON/XOFF chars come up
				*(out_buf+i) = rdata(8);
			} while( Xon && (out_buf[i] == XON_CHAR || out_buf[i] == XOFF_CHAR ) );
		}
	} else
		out_buf = pattern;

	if( (in_buf = malloc( patlen )) == NULL ) {
		fprintf( stderr, "Cross_Talk in_buf malloc failed, test #%ld\n", ltno );
		if( freeout )
			free( out_buf );
		return( 1 );
	}

	if( (flags=fcntl( out_fd, F_GETFL )) <  0
	|| fcntl( out_fd, F_SETFL, flags|O_NONBLOCK ) < 0 ) {
		fprintf( stderr,
			"%s: cannot set non-blocking mode on output port, %s\n",
			cmdname, strerror( errno ) );
		if( freeout )
			free( out_buf );
		return( 1 );
	}

	err=0;
	timeouts=0;
	write_chars = read_chars = 0;
	while( !stop_flag && !err
	&& (write_chars < patlen || read_chars < patlen) ) {

		if( write_chars < patlen ) {

			ret = write( out_fd, out_buf+write_chars, patlen-write_chars );

			if( ret >= 0 ) {
				write_chars += ret;
			}
			else if( errno != EAGAIN ) {
				err++;
				fprintf( stderr, "Write failed on %s, %s, test #%ld\n",
					out_ttyname, strerror( errno ), ltno );
			}

		}

		if( read_chars < patlen ) {

			ret = read( in_fd, in_buf+read_chars, patlen-read_chars );

			if( ret > 0 )
				read_chars += ret;
			else if( ret < 0 ) {
				++err;
				fprintf( stderr, "Read failed on %s, %s, test #%ld\n", in_ttyname,
					strerror( errno ), ltno );
			} else {
				fprintf( stderr, "Read timeout on %s after %d bytes, test #%ld\n", in_ttyname, read_chars,
					ltno );
				timeouts++;
				if( timeouts > MAXTIMEOUTS )
					++err;
			}
		}

		if( anykey_abort )
			(void)kbhit();

	}

	/*if( stop_flag )
	 *	err++;
	 *if( !err )
	 */
	if( !stop_flag && !err )
	{
		err = memcmp( out_buf, in_buf, patlen );
		if( err )
			fprintf( stderr,
			"Output and input patterns differ, test #%ld\n", ltno );
	}
	
	if( fcntl( out_fd, F_SETFL, flags ) < 0 ) {
		fprintf( stderr,
			"%s: cannot clear non-blocking mode on output port, %s\n",
			cmdname, strerror( errno ) );
	}

	if( freeout )	/* Only free output buffer if we created it here */
		free( out_buf );
	free( in_buf );
	return( err );

}

int Output_Only( long tno, int out_fd )
{
	int err, i, ret, nchars, freeout, timeouts, ltno;
	char *out_buf, *in_buf;

	ltno = tno+1;	/* For users start counting at 1 not 0 */

	/* If user did not specify a pattern, malloc space and
	 * generate a random pattern */
	freeout = 0;
	if( pattern == NULL ) {
		freeout = 1;
		if( (out_buf = malloc( patlen )) == NULL ) {
			fprintf( stderr, "Output_Only out_buf malloc failed, test #%ld\n",
			ltno );
			return( 1 );
		}
		for( i=0; i<patlen; ++i )
			*(out_buf+i) = rdata(8);
	} else
		out_buf = pattern;

	err=0;
	if( write( out_fd, out_buf, patlen ) != patlen ) {
			err++;
			fprintf( stderr, "Write failed on %s, %s, test #ld\n",
				out_ttyname, strerror( errno ), ltno );
	}

	if( stop_flag )
		err++;

	if( freeout )	/* Only free output buffer if we created it here */
		free( out_buf );
	return( err );

}

int get_tios( int *fd, char *ttyname, struct termios *tios ) {
	int ret;
	/* open tty */
	if( (*fd=open( ttyname, O_RDWR ) ) < 0 ) {
		fprintf( stderr, "%s: open failed on %s, %s\n",
			cmdname, ttyname, strerror( errno ) );
		return( 1 );
	}

	/* get stty settings */
	if( (ret=ioctl( *fd, TCGETS, (void *)tios )) < 0 ) {
		fprintf( stderr, "%s: TCGETS failed, ret=%d, %s\n", cmdname,
			ret, strerror( errno ) );
		return( 1 );
	}

	/* default values */
	tios->c_iflag &= ~(IXON|IXOFF|INLCR|ICRNL);
	tios->c_oflag &= ~(OPOST|ONLCR|OCRNL);
	tios->c_oflag |= ONLRET;
	tios->c_cflag &= ~(CRTSCTS|CBAUD|CSIZE|CSTOPB|PARENB);
	tios->c_cflag |= CLOCAL|CREAD|B9600|CS8;
	tios->c_lflag &=
		~(ICANON|ECHO|ECHOE|ECHOK|ECHOKE|ECHONL|ECHOCTL|ISIG|IEXTEN);
	// Make reads return immediately with data or after 1 second without
	tios->c_cc[VMIN] = 0;
	tios->c_cc[VTIME] = 10;	// 10ths of second
	return( 0 );
}

int set_tios( int *fd, char *ttyname, struct termios *tios ) {
	int ret;
	/* set stty settings */
	if( (ret=ioctl( *fd, TCSETS, (void *)tios )) < 0 ) {
		fprintf( stderr, "%s: TCSETS failed, ret=%d, %s\n", cmdname,
			ret, strerror( errno ) );
		return( 1 );
	}
	return( 0 );
}


void got_quit()
{
	printf( "Quit signal from user.\n" );
	stop_flag = 1;
}

/*
*  Catch Signal INT
*/

void catch_int( int sig )
{
	got_quit();
}


/*
*  Usage
*/

void usage(FILE *out)
{

	fprintf( out,
		"\nusage: %s [options] -i ttyname -o ttyname\n", cmdname );

	fprintf( out, "\nOptions:\n");
	fprintf( out, "        -v     Display version number and quit\n");
	fprintf( out, "        -h     Display this usage message\n");
	fprintf( out, "        -n #   Repetition count (how many times to run the tests)\n");
	fprintf( out, "               (A count of -1 specifies an infinite test loop)\n");
	fprintf( out, "        -B     Bidirectional -- perform tests in both directions\n");
	fprintf( out, "        -b #   Baudrate\n");
	fprintf( out, "        -S     Enable software flow control\n");
	fprintf( out, "        -H     Enable hardware flow control\n");
	fprintf( out, "        -e     Enable 'echo'\n");
	fprintf( out, "        -l #   Pattern length for random patterns\n");
	fprintf( out, "        -L     Enable internal UART loopback mode\n");
	fprintf( out, "        -m     Perform modem line tests after each data test\n");
	fprintf( out, "        -E     Some error messages go to stdout, with -E they will also be\n");
	fprintf( out, "               written to stderr (so stderr may be redirected to a file)\n");
	fprintf( out, "        -A     Allow any keypress to abort tests\n");
	fprintf( out, "        -R     Display summary report when finished\n");
	fprintf( out, "        -Ppattern\n");
	fprintf( out, "               By default the tests use random patterns but a particular\n");
	fprintf( out, "               pattern can be specified here in hex.  E.g:\n");
	fprintf( out, "                        -P0x4142430D0A\n");
	fprintf( out, "               specifies a 5 byte pattern corresponding to ASCII \"ABC\\r\\n\"\n");
	fprintf( out, "\n");
	fprintf( out, "       Notes:\n");
	fprintf( out, "       1. For modem (-m) tests, the input and output ports must be wired\n");
	fprintf( out, "          with signal and frame grounds straight through and:\n");
	fprintf( out, "          RxD  -- TxD          TxD     -- RxD\n");
	fprintf( out, "          RTS  -- CTS          CTS     -- RTS\n");
	fprintf( out, "          DTR  -- DCD/DSR      DCD/DSR -- DTR\n");
	fprintf( out, "       2. The input and output ports can be one and the same if connected\n");
	fprintf( out, "          with an external loopback plug wired as above.\n" );

}

