/* blast.c
 * Send user-specified or randomly generated data to a device.
 *
 * History:
 *     V0.1: Peter Berger (PEB): 17 January 2001
 *         -- First version
 *     V0.2: Peter Berger (PEB): 3 April 2003
 *         -- Added rascon.h reference
 */
/* includes */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <linux/types.h>
#include <curses.h>
#include <limits.h>

//#include <sys/types.h>
#include <sys/stat.h>

// CPCIDIR -- where to look for the driver header files
#include RASDIR"/rascon.h"
#include "utest.h"

/* typedefs */
typedef struct {
	long baud_num;
	int baud_flag;
} Baud_Table_Entry;


/* functions */
void catch_sigint( int sig );
void usage();
int get_tios( int *fd, char *ttyname, struct termios *tios );
int set_tios( int *fd, char *ttyname, struct termios *tios );
int Output_Only( long tno, int out_fd );

#define VERSION "0.1"
#define NAMELEN 128
#define BUF_SIZE 128

#define DEFAULT_BAUDRATE 9600
#define DEFAULT_CNT 1
#define DEFAULT_PATLEN 10

// Maximum number of read timeouts before a given test is aborted 
#define MAXTIMEOUTS 5


/* globals */
static char *cmdname;
static char in_ttyname[NAMELEN];
static char out_ttyname[NAMELEN];
static char *pattern=NULL;
static long baud=DEFAULT_BAUDRATE,
	cnt=DEFAULT_CNT,
	patlen=DEFAULT_PATLEN;
static char outbuf[BUFSIZ];	// We'll remap stdout using setvbuf()

#define DEVT_CHAR  1	// Regular character device (e.g: /dev/tty)
#define DEVT_BLOCK 2	// Block device requiring lseeks to beginning
#define DEVT_TERM  3	// Character device requiring termios settings
static devtype = DEVT_CHAR;

static char *version = VERSION" ""$Revision: 1.1 $";

static int stop_flag = 0;
static int bidirect = 0;	// Set if CrossTalk tests should go both ways

static Baud_Table_Entry translate_baud[] = {
	{ 0, B0 },
	{ 50, B50 },
	{ 75, B75 },
	{ 110, B110 },
	{ 134, B134 },
	{ 150, B150 },
	{ 200, B200 },
	{ 300, B300 },
	{ 600, B600 },
	{ 1200, B1200 },
	{ 1800, B1800 },
	{ 2400, B2400 },
	{ 4800, B4800 },
	{ 9600, B9600 },
	{ 19200, B19200 },
	{ 38400, B38400 },
	{ 57600, B57600 },
	{ 115200, B115200 },
	{ 230400, B230400 },
	{ 460800, B460800 },
	{ 921600, B921600 },
	{ -1L, 0 }
};

/* get_pattern
 * Convert user-specified pattern (a hex-coded ASCII string) to the
 * byte array and store it in malloc'd space pointed to by the global
 * pointer "pattern".
 *
 * Return 0 on success, 1 on error
 */
int get_pattern( char *cpattern )
{
	unsigned char *cpt, *ipt, bbuf[3], byte;
	int len, mlen;
int i;

	cpt = cpattern;	// Character (ASCII) input
	ipt = pattern;	// Integer array output
	len=strlen(cpt);

//printf( "in get_pattern:, cpt:'%s', len:%d\n", cpt, len );
	if( strncasecmp( cpt, "0x", 2 ) ) {
		fprintf( stderr, "get_pattern missing leading \"0x\".  Ignoring pattern spec.\n" );
		return( 1 );
	}
	cpt += 2;
	len -= 2;

	if( len <= 0 ) {
		fprintf( stderr, "get_pattern failed to parse pattern '%s'\n", cpattern );
		return( 1 );
	}

//printf( "in get_pattern after stripping:, cpt:'%s', len:%d\n", cpt, len );
		
	mlen = (len / 2);	// Two ASCII input chars per integer byte output
//printf( "mlen:%d\n", mlen );
	if( (pattern=malloc(mlen)) == NULL ) {
		fprintf( stderr, "get_pattern malloc failed\n" );
		return( 1 );
	}

	ipt=pattern;
	while( len >= 2 ) {
		strncpy( bbuf, cpt, 2 );
		byte = strtol( bbuf, NULL, 16 );
//printf( "byte:0x%x\n", byte );
		*ipt++ = byte;
		cpt += 2;
		len -= 2;
	}
	patlen=mlen;

//for( ipt=pattern, i=0; i<patlen; i++, ipt++ )
//	printf( "ipt:%d: 0x%x\n", i, *ipt );

	return( 0 );

}

void showversion()
{
	printf( "%s: Version  %s\n", cmdname, version );
}

show_result( FILE *out,long l, int ret, long fails ) 
{
	fprintf( out, "     Test #%ld", l+1 );
	if( cnt > 0 )
		fprintf( out,"/%ld", cnt );

 	fprintf( out, " '%s'. (%ld Pass/%ld Fail)\n",
		ret? "Failed" : "Passed", l+1-fails, fails );
}


/*
*  Main
*/

int main( int argc, char *argv[] )
{

	int ret, out_fd;
	long l, fails;
	char *ttyname,*pflag;
	Baud_Table_Entry *pbaud;
	struct termios in_tios, out_tios;
	struct sigaction sa;
	char *pattern_pt, *testname, devtype_id;


	/* get command name */
	cmdname = argv[0];
	--argc;
	++argv;

	/* catch SIGINT */
	sa.sa_handler = catch_sigint;
	sigemptyset( &sa.sa_mask );
	sa.sa_flags = 0;
	if( sigaction( SIGINT, &sa, NULL ) != 0
	|| sigaction( SIGTERM, &sa, NULL ) != 0 ) {
		fprintf( stderr, "%s: sigaction failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	/* check args */
	if( argc < 1 ) {
		usage( );
		return( 0 );
	}

	/* check for version option */
	if( strcmp( *argv, "-version" ) == 0
	|| strcmp( *argv, "--version" ) == 0 ) {
		showversion();
		return( 0 );
	}


	strcpy( in_ttyname, "" );
	strcpy( out_ttyname, "" );
	/* get options */
	while( argc > 0 ) {

		/* handle each flag */
		for( pflag=*argv; *pflag != '\0'; pflag++ ) {

			/* switch on flag */
			switch( *pflag ) {

			case '-':
				break;

			case '?':
			case 'h':
				usage();
				return(0);

			case 'v':
				showversion();
				return(0);

			case 'H':
				in_tios.c_cflag |= CRTSCTS;
				out_tios.c_cflag |= CRTSCTS;
				break;

			case 'S':
				in_tios.c_iflag |= (IXON|IXOFF);
				out_tios.c_iflag |= (IXON|IXOFF);
				break;

			case 'e':
				in_tios.c_lflag |= ECHO;
				out_tios.c_lflag |= ECHO;
				break;

			case 'o':			/* Output file name */
				if( *(pflag+1) != '\0' ) {
					strncpy( out_ttyname, pflag+1, NAMELEN-1 );
				} else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -o without output file name\n",
							cmdname );
						return( 1 );
					} else
						strncpy( out_ttyname, *++argv, NAMELEN-1 );
				}
				if( get_tios( &out_fd, out_ttyname, &out_tios ) ){
					fprintf( stderr, "%s: get_tios failed, %s\n",
					cmdname, strerror( errno ) );
					return( 1 );
				}
				break;

			case 'D':
				if( *(pflag+1) != '\0' ) {
					devtype_id = *(pflag+1);
					pflag += 1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -D without Device type\n",
							cmdname );
						return( 1 );
					}
					devtype_id = **++argv;
				}
				switch( toupper(devtype_id) ) {
					case 'B':
						devtype = DEVT_BLOCK;
						break;
					case 'C':
						devtype = DEVT_CHAR;
						break;
					case 'T':
						devtype = DEVT_TERM;
						break;
					default:
						fprintf( stderr, "%s: Bad device type '%c' -- using 'C'\n",
							cmdname, devtype_id );
							break;
				}
						
				break;

			case 'b':

				if( *(pflag+1) != '\0' ) {
					baud = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -b without baud rate\n",
							cmdname );
						return( 1 );
					}
					baud = atol( *++argv );
				}

				/* translate baud rate */
				for( pbaud=translate_baud; pbaud->baud_num != -1; pbaud++ ) {
					if( baud == pbaud->baud_num ) {
						in_tios.c_cflag &= ~CBAUD;
						in_tios.c_cflag |= pbaud->baud_flag;
						out_tios.c_cflag &= ~CBAUD;
						out_tios.c_cflag |= pbaud->baud_flag;
						break;
					}
				}

				/* baud rate not found */
				if( pbaud->baud_num == -1 ) {
					fprintf( stderr, "%s: unknown baud rate %d\n", cmdname,
						baud );
					return( 1 );
				}

				break;

			case 'n':
				if( *(pflag+1) != '\0' ) {
					cnt = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -c without repetition count\n",
							cmdname );
						return( 1 );
					}
					cnt = atol( *++argv );
				}
				break;

			case 'l':
				if( *(pflag+1) != '\0' ) {
					patlen = atol( pflag+1 );
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -l without pattern length\n",
							cmdname );
						return( 1 );
					}
					if( pattern ) {
						fprintf( stderr, "%s: pattern length set implicitly by -P option, ignoring -l option\n",
							cmdname );
							++argv;
					} else
						patlen = atol( *++argv );
				}
				break;

			case 'P':	/* Use this pattern instead of the random default */
				if( *(pflag+1) != '\0' ) {
					pattern_pt = pflag+1;
					pflag += strlen(pflag)-1;
				}
				else {
					if( --argc <= 0 ) {
						fprintf( stderr, "%s: -P without pattern\n",
							cmdname );
						return( 1 );
					}
					pattern_pt = *++argv;
				}
//printf( "pattern_pt:'%s'\n", pattern_pt );
				if( get_pattern( pattern_pt ) )
					pattern_pt = NULL;
				break;


			default:
				fprintf( stderr, "%s: bad flag '%c'\n", cmdname, *pflag );
				usage();
				return( 1 );
				break;

			}

		}

		--argc;
		++argv;

	}
//printf( "pattern_pt2: '%s'\n", pattern_pt );

	srand((unsigned)time(NULL));	// Seed the random number generator
	if( setvbuf( stdout,outbuf,_IOLBF, BUFSIZ ) ) {
		fprintf( stderr, "%s: setvbuf failed, %s\n",
			cmdname, strerror( errno ) );
		return( 1 );
	}

	if( devtype == DEVT_TERM ){
printf( "Calling set_tios...\n" );
			if( set_tios( &out_fd, out_ttyname, &out_tios ) ) {
				fprintf( stderr, "%s: get_tios failed, %s\n",
					cmdname, strerror( errno ) );
				return( 1 );
			}
	}


	testname="Output_Only";
	printf("\
%s specifications:\n\
     Using %d byte pattern of %s data\n",
	testname,
	patlen, pattern == NULL? "Random-generated" : "User-specified" );
	printf("\
     Number of test iterations to run: " );
	if( cnt > 0 )
		printf( "%ld\n", cnt );
	else
		printf( "infinite loop\n" );
	printf("\
     Output port: %s, Baudrate:%ld\n", out_ttyname, baud );

	fails=0;
	// If cnt is negative, loop forever
	for( l=0; !stop_flag && (cnt>0? l<cnt: 1); ++l )
	{
		ret=Output_Only( l, out_fd );
		if( ret )
			fails++;
		show_result( stdout,l,ret,fails );
		if( ret )	// Log failures to stderr
			show_result( stderr,l,ret,fails );


	}


	/* close port */
	close( out_fd );
	if( pattern )
		free( pattern );

	return( 0 );

}

/* rdata
 * Return a bits-size byte of random data
 */
unsigned char rdata( int bits )
{
	unsigned int mask, ret;

	mask = (1 << bits) - 1;
	return( (unsigned char)(rand() & mask ) );
}

ssize_t
writen(int fd, const void *vptr, size_t n)
{
	size_t nleft;
	ssize_t nwritten;
	const char *ptr;

	ptr = vptr;
	nleft = n;

	while (nleft > 0) {
		if ( (nwritten = write(fd, ptr, nleft)) <= 0)
			return(nwritten);
		nleft -= nwritten;
		ptr += nwritten;
	}
	return(n);
}

int Output_Only( long tno, int out_fd )
{
	int err, i, ret, nchars, freeout, timeouts;
	unsigned char *out_buf, *in_buf;
off_t seekret;

	/* If user did not specify a pattern, malloc space and
	 * generate a random pattern */
	freeout = 0;
	if( pattern == NULL ) {
		freeout = 1;
		if( (out_buf = malloc( patlen )) == NULL ) {
			fprintf( stderr, "Output_Only out_buf malloc failed, test #%ld\n",
			tno );
			return( 1 );
		}
		for( i=0; i<patlen; ++i )
			*(out_buf+i) = rdata(8);
	} else
		out_buf = pattern;

//printf( "Output_Only patlen:%d: %c %c %c\n", patlen, outbuf[0], outbuf[1], outbuf[2] );
	err=0;
	if( devtype == DEVT_BLOCK ) {
		if( (seekret=lseek( out_fd, (off_t)0, SEEK_SET )) == -1 ) {
			fprintf( stderr, "Seek failed on %s, %s, test #ld\n",
				out_ttyname, strerror( errno ), tno );
			err++;
		}
	}
	if( writen( out_fd, out_buf, patlen ) != patlen ) {
			err++;
			fprintf( stderr, "Write failed on %s, %s, test #%ld\n",
				out_ttyname, strerror( errno ), tno );
	}

	if( stop_flag )
		err++;

	if( freeout )	/* Only free output buffer if we created it here */
		free( out_buf );
	return( err );

}

int get_tios( int *fd, char *ttyname, struct termios *tios ) {
	int ret;
	int flags=O_RDWR;

	if( devtype == DEVT_BLOCK ){
		flags |= O_CREAT | O_TRUNC;
	}

	// Block devices may need to be created and truncated
	if( devtype == DEVT_BLOCK ) {
		if( (*fd=open( ttyname, O_WRONLY | O_CREAT | O_TRUNC | O_NONBLOCK,
		S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH)) < 0 ) {
			fprintf( stderr, "%s: open failed on %s, %s\n",
				cmdname, ttyname, strerror( errno ) );
			return( 1 );
		}
	} else if( (*fd=open( ttyname, O_WRONLY | O_NONBLOCK ) ) < 0 ) {
		fprintf( stderr, "%s: open failed on %s, %s\n",
			cmdname, ttyname, strerror( errno ) );
		return( 1 );
	}

	if( devtype == DEVT_TERM ){
		/* get stty settings */
		if( (ret=ioctl( *fd, TCGETS, (void *)tios )) < 0 ) {
			fprintf( stderr, "%s: TCGETS failed, ret=%d, %s\n", cmdname,
				ret, strerror( errno ) );
			return( 1 );
		}
	
		/* default values */
		tios->c_iflag &= ~(IXON|IXOFF|INLCR|ICRNL);
		tios->c_oflag &= ~(OPOST|ONLCR|OCRNL);
		tios->c_oflag |= ONLRET;
		tios->c_cflag &= ~(CRTSCTS|CBAUD|CSIZE|CSTOPB|PARENB);
		tios->c_cflag |= CLOCAL|CREAD|B9600|CS8;
		tios->c_lflag &=
			~(ICANON|ECHO|ECHOE|ECHOK|ECHOKE|ECHONL|ECHOCTL|ISIG|IEXTEN);
		// Make reads return immediately with data or after 1 second without
		tios->c_cc[VMIN] = 0;
		tios->c_cc[VTIME] = 10;	// 10ths of second
	}
	return( 0 );
}

int set_tios( int *fd, char *ttyname, struct termios *tios ) {
	int ret;
	/* set stty settings */
	if( (ret=ioctl( *fd, TCSETS, (void *)tios )) < 0 ) {
		fprintf( stderr, "%s: TCSETS failed, ret=%d, %s\n", cmdname,
			ret, strerror( errno ) );
		return( 1 );
	}
	return( 0 );
}


/*
*  Catch Signal INT
*/

void catch_sigint( int sig )
{
	stop_flag = 1;
//printf( "catch_sigint: caught sig:%d\n", sig );

}


/*
*  Usage
*/

void usage()
{

	fprintf( stderr,
		"\nusage: %s [options] -o ttyname\n", cmdname );

	fprintf( stderr, "\nOptions:\n\
        -v     Display version number and quit\n\
        -h     Display this usage message\n\
        -n #   Repetition count (how many times to run the tests)\n\
               (A count of -1 specifies an infinite test loop)\n\
        -b #   Baudrate\n\
        -S     Enable software flow control\n\
        -H     Enable hardware flow control\n\
        -e     Enable 'echo'\n\
        -l #   Pattern length for random patterns\n\
        -Db    Block device -- seek to beginning of file before each write\n\
        -Dt    Terminal device -- set termios parameters (baud,flow, etc.)\n\
        -Dc    Character device -- nothing special, just send output bytes\n\
        -Ppattern\n\
               By default the tests use random patterns but a particular\n\
               pattern can be specified here in hex.  E.g:\n\
                        -P0x4142430D0A\n\
               specifies a 5 byte pattern corresponding to ASCII \"ABC\\r\\n\"\n");


}
